@@ -64,7 +64,7 @@
 #endif
 #include "../gnome-pty-helper/gnome-pty.h"
 static gboolean _vte_pty_helper_started = FALSE;
-static GPid _vte_pty_helper_pid = -1;
+static pid_t _vte_pty_helper_pid = -1;
 static int _vte_pty_helper_tunnel = -1;
 static GTree *_vte_pty_helper_map = NULL;
 #endif
@@ -159,7 +159,7 @@ struct vte_pty_child_setup_data {
 		int fd;
 	} tty;
 };
-static void 
+static void
 vte_pty_child_setup (gpointer arg)
 {
 	struct vte_pty_child_setup_data *data = arg;
@@ -179,21 +179,14 @@ vte_pty_child_setup (gpointer arg)
 			break;
 	}
 
-#ifdef VTE_DEBUG
-	if (_vte_debug_on (VTE_DEBUG_PTY)) {
-		g_printerr ("Setting up child pty: name = %s, fd = %d\n",
+	_vte_debug_print (VTE_DEBUG_PTY,
+			"Setting up child pty: name = %s, fd = %d\n",
 				tty ? tty : "(none)", fd);
-	}
-#endif
 
 
 	/* Start a new session and become process-group leader. */
 #if defined(HAVE_SETSID) && defined(HAVE_SETPGID)
-#ifdef VTE_DEBUG
-	if (_vte_debug_on (VTE_DEBUG_PTY)) {
-		g_printerr ("Starting new session\n");
-	}
-#endif
+	_vte_debug_print (VTE_DEBUG_PTY, "Starting new session\n");
 	setsid();
 	setpgid(0, 0);
 #endif
@@ -221,11 +214,11 @@ vte_pty_child_setup (gpointer arg)
 	if ((ioctl(fd, I_FIND, "ptem") == 0) &&
 			(ioctl(fd, I_PUSH, "ptem") == -1)) {
 		_exit (127);
-	} 
+	}
 	if ((ioctl(fd, I_FIND, "ldterm") == 0) &&
 			(ioctl(fd, I_PUSH, "ldterm") == -1)) {
 		_exit (127);
-	} 
+	}
 	if ((ioctl(fd, I_FIND, "ttcompat") == 0) &&
 			(ioctl(fd, I_PUSH, "ttcompat") == -1)) {
 		perror ("ioctl (fd, I_PUSH, \"ttcompat\")");
@@ -264,6 +257,11 @@ vte_pty_child_setup (gpointer arg)
 	_vte_pty_reset_signal_handlers();
 }
 
+/* TODO: clean up the spawning
+ * - replace current env rather than adding!
+ * - allow user control over flags (eg DO_NOT_CLOSE)
+ * - additional user callback for child setup
+ */
 
 /* Run the given command (if specified) */
 static gboolean
@@ -273,42 +271,83 @@ _vte_pty_run_on_pty (struct vte_pty_child_setup_data *data,
 		     GPid *pid, GError **error)
 {
 	gboolean ret = TRUE;
+	GError *local_error = NULL;
 
 	if (command != NULL) {
-		gchar **arg2;
-		gint i, argc;
+		gchar **arg2, **envp2;
+		gint i, k, argc;
 
 		/* push the command into argv[0] */
 		argc = argv ? g_strv_length (argv) : 0;
-		arg2 = g_new (char *, argc+2);
+		arg2 = g_new (char *, argc + 2);
 		arg2[0] = g_strdup (command);
-#ifdef VTE_DEBUG
-		if (_vte_debug_on (VTE_DEBUG_MISC)) {
-			g_printerr("Spawing command...\nargv[0] = %s\n",
-				       	arg2[0]);
-		}
-#endif
-		for (i=0; i<argc; i++) {
+		for (i = 0; i < argc; i++) {
 			arg2[i+1] = g_strdup (argv[i]);
-#ifdef VTE_DEBUG
-			if (_vte_debug_on (VTE_DEBUG_MISC)) {
-				g_printerr("argv[%d] = %s\n", i+1, arg2[i+1]);
-			}
-#endif
 		}
 		arg2[i+1] = NULL;
 
+		/* add the given environment to the childs */
+		i = g_strv_length (environ) + (envp ? g_strv_length (envp) : 0);
+		envp2 = g_new (char *, i + 1);
+		for (i = 0; environ[i] != NULL; i++) {
+			envp2[i] = g_strdup (environ[i]);
+		}
+		if (envp != NULL) {
+			for (k = 0; envp[k] != NULL; k++) {
+				envp2[i++] = g_strdup (envp[k]);
+			}
+		}
+		envp2[i] = NULL;
+
+		_VTE_DEBUG_IF (VTE_DEBUG_MISC) {
+			g_printerr ("Spawing command '%s'\n", command);
+			for (i = 0; arg2[i] != NULL; i++) {
+				g_printerr ("    argv[%d] = %s\n", i, arg2[i]);
+			}
+			for (i = 0; envp2[i] != NULL; i++) {
+				g_printerr ("    env[%d] = %s\n", i, envp2[i]);
+			}
+			g_printerr ("    directory: %s\n",
+					directory ? directory : "(none)");
+		}
+
 		ret = g_spawn_async_with_pipes (directory,
-				arg2, envp,
+				arg2, envp2,
 				G_SPAWN_CHILD_INHERITS_STDIN |
-				G_SPAWN_FILE_AND_ARGV_ZERO |
 				G_SPAWN_SEARCH_PATH |
-				G_SPAWN_DO_NOT_REAP_CHILD,
+				G_SPAWN_DO_NOT_REAP_CHILD |
+				(argv ? G_SPAWN_FILE_AND_ARGV_ZERO : 0),
 				vte_pty_child_setup, data,
 				pid,
 				NULL, NULL, NULL,
-				error);
+				&local_error);
+		if (ret == FALSE) {
+			if (g_error_matches (local_error,
+						G_SPAWN_ERROR,
+						G_SPAWN_ERROR_CHDIR)) {
+				/* try spawning in our working directory */
+				g_clear_error (&local_error);
+				ret = g_spawn_async_with_pipes (NULL,
+						arg2, envp2,
+						G_SPAWN_CHILD_INHERITS_STDIN |
+						G_SPAWN_SEARCH_PATH |
+						G_SPAWN_DO_NOT_REAP_CHILD |
+						(argv ? G_SPAWN_FILE_AND_ARGV_ZERO : 0),
+						vte_pty_child_setup, data,
+						pid,
+						NULL, NULL, NULL,
+						&local_error);
+			}
+		}
 		g_strfreev (arg2);
+		g_strfreev (envp2);
+
+		_vte_debug_print (VTE_DEBUG_MISC,
+				"Spawn result: %s%s\n",
+				ret?"Success":"Failure - ",
+				ret?"":local_error->message);
+		if (local_error)
+			g_propagate_error (error, local_error);
 	}
 #ifdef HAVE_FORK
 	else {
@@ -400,21 +439,15 @@ _vte_pty_set_size(int master, int columns, int rows)
 	memset(&size, 0, sizeof(size));
 	size.ws_row = rows ? rows : 24;
 	size.ws_col = columns ? columns : 80;
-#ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_PTY)) {
-		g_printerr("Setting size on fd %d to (%d,%d).\n",
+	_vte_debug_print(VTE_DEBUG_PTY,
+			"Setting size on fd %d to (%d,%d).\n",
 			master, columns, rows);
-	}
-#endif
 	ret = ioctl(master, TIOCSWINSZ, &size);
-#ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_PTY)) {
-		if (ret != 0) {
-			g_printerr("Failed to set size on %d: %s.\n",
+	if (ret != 0) {
+		_vte_debug_print(VTE_DEBUG_PTY,
+				"Failed to set size on %d: %s.\n",
 				master, strerror(errno));
-		}
 	}
-#endif
 	return ret;
 }
 
@@ -442,19 +475,13 @@ _vte_pty_get_size(int master, int *columns, int *rows)
 		if (rows != NULL) {
 			*rows = size.ws_row;
 		}
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("Size on fd %d is (%d,%d).\n",
+		_vte_debug_print(VTE_DEBUG_PTY,
+				"Size on fd %d is (%d,%d).\n",
 				master, size.ws_col, size.ws_row);
-		}
-#endif
 	} else {
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("Failed to read size from fd %d.\n",
+		_vte_debug_print(VTE_DEBUG_PTY,
+				"Failed to read size from fd %d.\n",
 				master);
-		}
-#endif
 	}
 	return ret;
 }
@@ -472,11 +499,8 @@ _vte_pty_ptsname(int master)
 		switch (i) {
 		case 0:
 			/* Return the allocated buffer with the name in it. */
-#ifdef VTE_DEBUG
-			if (_vte_debug_on(VTE_DEBUG_PTY)) {
-				g_printerr("PTY slave is `%s'.\n", buf);
-			}
-#endif
+			_vte_debug_print(VTE_DEBUG_PTY,
+					"PTY slave is `%s'.\n", buf);
 			return buf;
 			break;
 		default:
@@ -489,21 +513,14 @@ _vte_pty_ptsname(int master)
 #elif defined(HAVE_PTSNAME)
 	char *p;
 	if ((p = ptsname(master)) != NULL) {
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("PTY slave is `%s'.\n", p);
-		}
-#endif
+		_vte_debug_print(VTE_DEBUG_PTY, "PTY slave is `%s'.\n", p);
 		return g_strdup(p);
 	}
 #elif defined(TIOCGPTN)
 	int pty = 0;
 	if (ioctl(master, TIOCGPTN, &pty) == 0) {
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("PTY slave is `/dev/pts/%d'.\n", pty);
-		}
-#endif
+		_vte_debug_print(VTE_DEBUG_PTY,
+				"PTY slave is `/dev/pts/%d'.\n", pty);
 		return g_strdup_printf("/dev/pts/%d", pty);
 	}
 #endif
@@ -564,27 +581,20 @@ _vte_pty_open_unix98(GPid *child, char **env_add,
 
 	/* Attempt to open the master. */
 	fd = _vte_pty_getpt();
-#ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_PTY)) {
-		g_printerr("Allocated pty on fd %d.\n", fd);
-	}
-#endif
+	_vte_debug_print(VTE_DEBUG_PTY, "Allocated pty on fd %d.\n", fd);
 	if (fd != -1) {
 		/* Read the slave number and unlock it. */
 		if (((buf = _vte_pty_ptsname(fd)) == NULL) ||
 		    (_vte_pty_grantpt(fd) != 0) ||
 		    (_vte_pty_unlockpt(fd) != 0)) {
-#ifdef VTE_DEBUG
-			if (_vte_debug_on(VTE_DEBUG_PTY)) {
-				g_printerr("PTY setup failed, bailing.\n");
-			}
-#endif
+			_vte_debug_print(VTE_DEBUG_PTY,
+					"PTY setup failed, bailing.\n");
 			close(fd);
 			fd = -1;
 		} else {
 			/* Start up a child process with the given command. */
 			if (!_vte_pty_fork_on_pty_name(buf, fd,
-					       	      env_add, command,
+						      env_add, command,
 						      argv, directory,
 						      columns, rows,
 						      child)) {
@@ -858,39 +868,24 @@ _vte_pty_open_with_helper(GPid *child, char **env_add,
 			    &ops, sizeof(ops)) != sizeof(ops)) {
 			return -1;
 		}
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("Sent request to helper.\n");
-		}
-#endif
+		_vte_debug_print(VTE_DEBUG_PTY, "Sent request to helper.\n");
 		/* Read back the response. */
 		if (n_read(_vte_pty_helper_tunnel,
 			   &ret, sizeof(ret)) != sizeof(ret)) {
 			return -1;
 		}
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("Received response from helper.\n");
-		}
-#endif
+		_vte_debug_print(VTE_DEBUG_PTY,
+				"Received response from helper.\n");
 		if (ret == 0) {
 			return -1;
 		}
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("Helper returns success.\n");
-		}
-#endif
+		_vte_debug_print(VTE_DEBUG_PTY, "Helper returns success.\n");
 		/* Read back a tag. */
 		if (n_read(_vte_pty_helper_tunnel,
 			   &tag, sizeof(tag)) != sizeof(tag)) {
 			return -1;
 		}
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("Tag = %p.\n", tag);
-		}
-#endif
+		_vte_debug_print(VTE_DEBUG_PTY, "Tag = %p.\n", tag);
 		/* Receive the master and slave ptys. */
 		_vte_pty_read_ptypair(_vte_pty_helper_tunnel,
 				      &parentfd, &childfd);
@@ -900,12 +895,9 @@ _vte_pty_open_with_helper(GPid *child, char **env_add,
 			close(childfd);
 			return -1;
 		}
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PTY)) {
-			g_printerr("Got master pty %d and slave pty %d.\n",
+		_vte_debug_print(VTE_DEBUG_PTY,
+				"Got master pty %d and slave pty %d.\n",
 				parentfd, childfd);
-		}
-#endif
 
 		/* Add the parent and the tag to our map. */
 		g_tree_insert(_vte_pty_helper_map,
@@ -948,11 +940,12 @@ _vte_pty_open_with_helper(GPid *child, char **env_add,
  * Returns: an open file descriptor for the pty master, -1 on failure
  */
 int
-_vte_pty_open(GPid *child, char **env_add,
+_vte_pty_open(pid_t *child_pid, char **env_add,
 	      const char *command, char **argv, const char *directory,
 	      int columns, int rows,
 	      gboolean lastlog, gboolean utmp, gboolean wtmp)
 {
+	GPid child;
 	int ret = -1;
 #ifdef VTE_USE_GNOME_PTY_HELPER
 	int op = 0;
@@ -978,20 +971,21 @@ _vte_pty_open(GPid *child, char **env_add,
 	g_assert(op >= 0);
 	g_assert(op < G_N_ELEMENTS(opmap));
 	if (ret == -1) {
-		ret = _vte_pty_open_with_helper(child, env_add, command, argv,
+		ret = _vte_pty_open_with_helper(&child, env_add, command, argv,
 						directory,
 						columns, rows, opmap[op]);
 	}
 #endif
 	if (ret == -1) {
-		ret = _vte_pty_open_unix98(child, env_add, command, argv,
+		ret = _vte_pty_open_unix98(&child, env_add, command, argv,
 					   directory, columns, rows);
 	}
-#ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_PTY)) {
-		g_printerr("Returning ptyfd = %d.\n", ret);
+	if (ret != -1) {
+		*child_pid = (pid_t) child;
 	}
-#endif
+	_vte_debug_print(VTE_DEBUG_PTY,
+			"Returning ptyfd = %d, child = %ld.\n",
+			ret, (long) child);
 	return ret;
 }
 
@@ -1007,7 +1001,7 @@ _vte_pty_open(GPid *child, char **env_add,
 void
 _vte_pty_set_utf8(int pty, gboolean utf8)
 {
-#ifdef IUTF8
+#if defined(HAVE_TCSETATTR) && defined(IUTF8)
 	struct termios tio;
 	tcflag_t saved_cflag;
 	if (pty != -1) {
